unit global;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, IniFiles, streamex, DateUtils, Character, Math, memds, DB, DCPsha256;

type

  { TSettings }

  TSettings = class
  private
    fini: TIniFile;
  public
    constructor Create;
    destructor Destroy; override;
    function sectionValue(ASection, AIndent: string): string;
  end;

  { TDelta }

  TDelta = class
  private
    fDebugOption: integer;
    fDeltaFile: string;
    fNewFile: string;
    fOldFile: string;
    fstart: TTime;
    fend: TTime;
  private
    vSettings: TSettings;
    fcurExecTime: TTime;
    fexec: boolean;
    fNewDataSet, fOldDataSet: TMemDataset;
    procedure startCompare(showDebugOption: integer);
    function compareLine(AText: string; AIndex: integer; var AEnd: TTime): boolean;
    procedure compareStart;
    procedure createDeltaFile(aText: string);
    property OldFile: string read fOldFile;
    property NewFile: string read fNewFile;
    property DeltaFile: string read fDeltaFile;
    property DebugOption: integer read fDebugOption write fDebugOption;
    procedure CreateDataSet;
    procedure AddToDataSet(AFile: string; IsNew: boolean);
    procedure WriteFileDetails;
    function getTotalRows(isNew: boolean): integer;
    procedure compareDelta;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Execute(ANewFile, AOldFile, ADeltaFile: string; showDebugOption: integer);
  end;

function StrIsInteger(const S: string): boolean;

const
  VERSION = '1.0.2';
  APP = 'Claim Delta Compare';

implementation

const
  PROCESS = 'Process completed. Started %s, Ended %s, ' + 'Elapsed Time (seconds) %s';
  {$IFDEF MSWINDOWS}
  CRLF = #13#10;
  {$ELSE}
  CRLF = #10;

  {$ENDIF}


function getsha256(S: string): string;
var
  Hash: TDCP_sha256;
  Digest: array[0..31] of byte;  // sha256 produces a 256bit digest (32bytes)
  Source: string;
  i: integer;
  str1: string;
begin
  Source := S;  // here your string for get sha256

  if Source <> '' then
  begin
    Hash := TDCP_sha256.Create(nil);  // create the hash
    Hash.Init;                        // initialize it
    Hash.UpdateStr(Source);
    Hash.Final(Digest);               // produce the digest
    str1 := '';
    for i := 0 to 31 do
      str1 := str1 + IntToHex(Digest[i], 2);
    Result := UpperCase(str1);
  end;
end;

function CharIsDigit(const C: char): boolean;
begin
  Result := TCharacter.IsDigit(C);
end;


function CharIsNumber(const C: char): boolean;
begin
  Result := CharIsDigit(C);
end;

function StrIsInteger(const S: string): boolean;
var
  I: integer;
  Ch: char;
begin
  Result := S <> '';
  for I := 1 to Length(S) do
  begin
    Ch := S[I];
    if (not CharIsNumber(Ch)) then //Az
    begin
      Result := False;
      Exit;
    end;
  end;
end;

function CompareStreams(Stream1, Stream2: TStream): extended;
type
  TCompareBuffer = array[0..8191] of byte;

const
  // store the buffer's size
  szCompareBuffer = SizeOf(TCompareBuffer);

var
  // buffer variables, one for each stream
  Buffer1: TCompareBuffer;
  Buffer2: TCompareBuffer;
  // variables that will store the actual read bytes from streams
  ReadBytes1: integer;
  ReadBytes2: integer;
  // declare a variable that will store the number of different bytes in streams
  DifferenceCount: int64;
  // loop variable
  Index: integer;
  // max difference check loop's per buffer
  MaxCount: integer;
begin
  // set stream position to 0
  Stream1.Position := 0;
  Stream2.Position := 0;
  // initialize difference count
  DifferenceCount := 0;
  // start a loop
  while True do
  begin
    // read from both streams
    ReadBytes1 := Stream1.Read(Buffer1, szCompareBuffer);
    ReadBytes2 := Stream2.Read(Buffer2, szCompareBuffer);
    // set the max count to the smaller value of read bytes
    MaxCount := Min(ReadBytes1, ReadBytes2);
    // check differences byte by byte
    for Index := 0 to MaxCount - 1 do
      if Buffer1[Index] <> Buffer2[Index] then
        // difference found! increment DifferenceCount variable
        Inc(DifferenceCount);
    // if the number of read bytes from Stream1 is different than the
    // number of read bytes from Stream or we haven't read any bytes from
    // a stream, then break the loop, we're done comparing
    if (ReadBytes1 <> ReadBytes2) or (ReadBytes1 = 0) or (ReadBytes2 = 0) then
      Break;
  end; // while True do begin
  // return the number of differences
  Result := (DifferenceCount * 100) / Max(Stream1.Size, Stream2.Size);
end;

{ TDelta }

procedure TDelta.startCompare(showDebugOption: integer);
var
  fSourceLine: string;
  fs, fswriter: TFileStream;
  stream: TStreamReader;
  i, j: integer;
  {vstart, }vend: TTime;
begin
  i := 0;
  fs := TFileStream.Create(NewFile, fmOpenRead or fmShareDenyRead);
  stream := TStreamReader.Create(fs);
  fexec := False;
  try
    while not stream.EOF do
    begin
      fSourceLine := stream.ReadLine;
      if compareLine(fSourceLine, i, vend) then
      begin
        case showDebugOption of
          1:
          begin
            if i mod 1000 = 0 then
            begin
              Write(Format('Comparing Row %d ', [i + 1]));
              WriteLn(' Base Start ', FormatDateTime('hh:mm:ss:ms', fstart), ' Base elapsed ', FormatDateTime('hh:mm:ss:ms', vend - fstart),
                ' Start at ', FormatDateTime('hh:mm:ss:ms', fcurExecTime), ' end at ',
                FormatDateTime('hh:mm:ss:ms', vend),
                ' time elapsed for checking ', FormatDateTime('hh:mm:ss:ms', vend - fcurExecTime));
            end;
          end;

          2:
          begin
            Write(Format('Comparing Row %d ', [i + 1]));
            WriteLn(' Base Start ', FormatDateTime('hh:mm:ss:ms', fstart), ' Base elapsed ', FormatDateTime('hh:mm:ss:ms', vend - fstart),
              ' Start at ', FormatDateTime('hh:mm:ss:ms', fcurExecTime), ' end at ',
              FormatDateTime('hh:mm:ss:ms', vend),
              ' time elapsed for checking ', FormatDateTime('hh:mm:ss:ms', vend - fcurExecTime));
          end;

          3:
          begin
            Write(Format('Comparing Row %d ', [i + 1]));
            WriteLn(' Base Start ', FormatDateTime('hh:mm:ss:ms', fstart), ' Base elapsed ', FormatDateTime('hh:mm:ss:ms', vend - fstart),
              ' Start at ', FormatDateTime('hh:mm:ss:ms', fcurExecTime), ' end at ',
              FormatDateTime('hh:mm:ss:ms', vend),
              ' time elapsed for checking ', FormatDateTime('hh:mm:ss:ms', vend - fcurExecTime), #13#10, 'Line %s ', fSourceLine);
          end;
        end;
        fSourceLine := fSourceLine + CRLF;
        if (FileExists(DeltaFile)) then
        begin
          fswriter := TFileStream.Create(DeltaFile, fmOpenWrite);// or fmShareDenyNone);
          try
            fswriter.GoToEnd;
            for j := 1 to Length(fSourceLine) do
              fswriter.Write(fSourceLine[j], Length(fSourceLine[j]));
          finally
            fswriter.Free;
          end;
        end
        else
        begin
          fswriter := TFileStream.Create(DeltaFile, fmCreate);
          try
            for j := 1 to Length(fSourceLine) do
              fswriter.Write(fSourceLine[j], Length(fSourceLine[j]));
          finally
            fswriter.Free;
          end;
        end;
        //vend := Now;
      end
      else
        Continue;
      fexec := True;
      fcurExecTime := vend;
      Inc(i);
    end;
  finally
    stream.BaseStream.Free;
    stream.Close;
    stream.Free;
    fend := Now;
  end;
  Writeln(Format(PROCESS, [FormatDateTime('mm-dd-yyyy hh:mm:ss AMPM', fstart), FormatDateTime('mm-dd-yyyy hh:mm:ss AMPM', fend),
    FormatDateTime('hh:mm:ss:ms', fEnd - fStart)]));
  Exit;
end;

function TDelta.compareLine(AText: string; AIndex: integer; var AEnd: TTime): boolean;
var
  //fSourceLine: string;
  fs, cs: TFileStream;
  //stream: TStreamReader;
  i: integer;
begin
  Result := False;
  i := 0;
  //j := 0;
  if not fexec then
    fcurExecTime := Now;
  fs := TFileStream.Create(OldFile, fmOpenReadWrite or fmShareDenyNone);
  cs := TFileStream.Create(ExtractFilePath(ParamStr(0)) + 'temp.csv', fmCreate);
  try
    for i := 1 to Length(AText) do
      cs.Write(AText[i], Length(AText[i]));
    Result := CompareStreams(cs, fs) = 0;
    AEnd := Now;
  finally
    fs.Free;
    cs.Free;
    DeleteFile(ExtractFilePath(ParamStr(0)) + 'temp.csv');
  end;

  (*Exit;
  stream := TStreamReader.Create(fs);
  try
    while not stream.EOF do
    begin
      fSourceLine := stream.ReadLine;
      Result := (not SameText(fSourceLine, AText)) and (i = AIndex);
      if Result then
      begin
        AEnd := Now;
        Break;
      end;
      {
      if CompareText(fSourceLine, AText) <> 0 then
      begin
        Inc(j);
        //Result := True;
        //if i = AIndex then
          //break;
      end
      else
        Continue;
      }
      Inc(i);
    end;
  finally
    stream.BaseStream.Free;
    stream.Close;
    stream.Free;
  end;
  *)
  //Result := j < i;
  //AEnd := Now;
end;

procedure TDelta.compareStart;
var
  i: integer;
  {vcommit, }veach: boolean;
  ftext: string;
  fs: TFileStream;
  stream: TStreamReader;
  vcend: TTime;
begin
  i := 0;
  fs := TFileStream.Create(NewFile, fmOpenRead or fmShareDenyNone);
  stream := TStreamReader.Create(fs);
  fexec := False;
  try
    while not stream.EOF do
    begin
      ftext := stream.ReadLine;
      {
      if i = 0 then
        vcommit := False
      else
        vcommit := i mod 2000 = 0;
      }
      case DebugOption of
        1: veach := i mod 1000 = 0;
        2, 3: veach := i mod 1 = 0;
      end;

      //if vsql <> '' then
      //executeSQL(vsql, vcommit, False, i);
      if compareLine(ftext, i, vcend) then
      begin
        ftext := ftext + CRLF;
        createDeltaFile(ftext);
      end;
      if veach then
      begin
        //vcend := Now;
        Write('Comparing row no. : ' + (i + 1).ToString + ' ');
        WriteLn('Base Start ', FormatDateTime('hh:mm:ss:ms', fstart), ' Base elapsed ', FormatDateTime('hh:mm:ss:ms', vcend - fstart),
          ' Started: ' + FormatDateTime('hh:mm:ss:ms AMPM', fcurExecTime), ' Ended: ' + FormatDateTime('hh:mm:ss:ms AMPM', vcend),
          ' in ', FormatDateTime('hh:mm:ss:ms', vcend - fcurExecTime));
      end;
      fexec := True;
      fcurExecTime := vcend;
      Inc(i);
    end;
  finally
    stream.BaseStream.Free;
    stream.Close;
    stream.Free;
  end;
  fEnd := Now;
  WriteLn('Execution completed at ', FormatDateTime('hh:mm:ss AMPM', fEnd), ' Started at ', fStart);
  Write('Time elapsed: ');
  Writeln(FormatDateTime('hh:mm:ss:ms', fEnd - fStart));
end;

procedure TDelta.createDeltaFile(aText: string);
var
  j: integer;
  fswriter: TFileStream;
begin
  if (FileExists(DeltaFile)) then
  begin
    fswriter := TFileStream.Create(DeltaFile, fmOpenWrite);// or fmShareDenyNone);
    try
      fswriter.GoToEnd;
      for j := 1 to Length(aText) do
        fswriter.Write(aText[j], Length(aText[j]));
    finally
      fswriter.Free;
    end;
  end
  else
  begin
    fswriter := TFileStream.Create(DeltaFile, fmCreate);
    try
      for j := 1 to Length(aText) do
        fswriter.Write(aText[j], Length(aText[j]));
    finally
      fswriter.Free;
    end;
  end;
end;

procedure TDelta.CreateDataSet;
begin
  with fNewDataSet.FieldDefs do
  begin
    Clear;
    Add('content', ftBlob);
    Add('hash', ftString, 300);
  end;
  fNewDataSet.CreateTable;
  with fOldDataSet.FieldDefs do
  begin
    Clear;
    Add('content', ftBlob);
    Add('hash', ftString, 300);
  end;
  fOldDataSet.CreateTable;
end;

procedure TDelta.AddToDataSet(AFile: string; IsNew: boolean);
var
  i: integer;
  {vcommit, }veach: boolean;
  ftext: string;
  fs: TFileStream;
  stream: TStreamReader;
  //vcend: TTime;
begin
  i := 0;
  fs := TFileStream.Create(AFile, fmOpenRead or fmShareDenyNone);
  stream := TStreamReader.Create(fs);
  fexec := False;
  try
    while not stream.EOF do
    begin
      ftext := stream.ReadLine;
      case DebugOption of
        1: veach := i mod 1000 = 0;
        2, 3: veach := i mod 1 = 0;
      end;

      if veach then
      begin
        //vcend := Now;
        if IsNew then
          WriteLn('Processing New File row no. : ' + (i + 1).ToString + ' ')
        else
          WriteLn('Processing Old File row no. : ' + (i + 1).ToString + ' ');
        //WriteLn(ftext.substring(0, 100));
        {
        WriteLn('Base Start ', FormatDateTime('hh:mm:ss:ms', fstart), ' Base elapsed ', FormatDateTime('hh:mm:ss:ms', vcend - fstart),
          ' Started: ' + FormatDateTime('hh:mm:ss:ms AMPM', fcurExecTime), ' Ended: ' + FormatDateTime('hh:mm:ss:ms AMPM', vcend),
          ' in ', FormatDateTime('hh:mm:ss:ms', vcend - fcurExecTime));
        }
      end;
      if IsNew then
      begin
        fNewDataSet.Append;
        fNewDataSet.FieldByName('content').AsString := ftext;
        fNewDataSet.FieldByName('hash').AsString := getsha256(ftext);
      end
      else
      begin
        fOldDataSet.Append;
        fOldDataSet.FieldByName('content').AsString := ftext;
        fOldDataSet.FieldByName('hash').AsString := getsha256(ftext);
      end;
      Inc(i);
    end;
  finally
    stream.BaseStream.Free;
    stream.Close;
    stream.Free;
    if IsNew then
      WriteLn('New File Process complete')
    else
      WriteLn('Old File Process complete')
  end;
end;

procedure TDelta.WriteFileDetails;
begin
  WriteLn('New file has ' + fNewDataSet.RecordCount.ToString + ' lines');
  WriteLn('Old file has ' + fOldDataSet.RecordCount.ToString + ' lines');
  WriteLn('Total rows ' + fNewDataSet.RecordCount.ToString);
end;

function TDelta.getTotalRows(isNew: boolean): integer;
var
  i: integer;
begin
  i := 0;
  if isNew then
  begin
    with fNewDataSet do
    begin
      First;
      DisableControls;
      while not EOF do
      begin
        if isNew then
        begin
          if FieldByName('isnew').AsInteger = 1 then
            Inc(I);
        end
        else
        begin
          if FieldByName('isold').AsInteger = 1 then
            Inc(I);
        end;
        Next;
      end;
      EnableControls;
      First;
    end;
  end;
  Result := i;
end;

procedure TDelta.compareDelta;
var
  i, j, k: integer;
  //veach: boolean;
  ftext: string;
  vcend: TTime;
begin
  i := 0;
  j := 0;
  k := 0;
  WriteLn('Starting comparing, it will take some time to complete. Please wait...');
  with fNewDataSet do
  begin
    First;
    DisableControls;
    while not EOF do
    begin
      {case DebugOption of
        1: veach := i mod 1000 = 0;
        2, 3: veach := i mod 1 = 0;
      end;}
      ftext := FieldByName('content').AsString;
      if fOldDataSet.Locate('content', ftext, [loCaseInsensitive]) then
      begin
        Inc(j);
        case DebugOption of
          1:
          begin
            if i mod 1000 = 0 then
              WriteLn('Match row, skipping ', (i + 1).ToString);
          end;
          2, 3:
          begin
            if i mod 1 = 0 then
              WriteLn('Match row, skipping ', (i + 1).ToString);
          end;
        end;
        //if veach then
        //WriteLn('Match row, skipping ', (i + 1).ToString);
      end
      else
      begin
        Inc(k);
        fexec := True;
        fcurExecTime := vcend;
        createDeltaFile(ftext);
        case DebugOption of
          1:
          begin
            if i mod 1000 = 0 then
              WriteLn('Non match row, adding ', (i + 1).ToString);
          end;
          2, 3:
          begin
            if i mod 1 = 0 then
              WriteLn('Non match row, adding ', (i + 1).ToString);
          end;
        end;
        ftext := ftext + CRLF;
        //if veach then
        //begin
        //Write('Comparing row no. : ' + (i + 1).ToString + ' ');
        {
        Write('Non match row, adding ', (i + 1).ToString);
        WriteLn('Base Start ', FormatDateTime('hh:mm:ss:ms', fstart), ' Base elapsed ', FormatDateTime('hh:mm:ss:ms', vcend - fstart),
          ' Started: ' + FormatDateTime('hh:mm:ss:ms AMPM', fcurExecTime), ' Ended: ' + FormatDateTime('hh:mm:ss:ms AMPM', vcend),
          ' in ', FormatDateTime('hh:mm:ss:ms', vcend - fcurExecTime));
        }
        //end;
      end;
      Inc(I);
      Next;
    end;
    EnableControls;
    First;
  end;
  fEnd := Now;
  WriteLn('Execution completed at ', FormatDateTime('hh:mm:ss AMPM', fEnd), ' Started at ', FormatDateTime('hh:mm:ss AMPM', fstart));
  WriteLn('Skipped rows ', j.ToString, ' out of ', fNewDataSet.RecordCount);
  WriteLn('Delta rows ', k.ToString, ' out of ', fNewDataSet.RecordCount);
  Write('Time elapsed: ');
  Writeln(FormatDateTime('hh:mm:ss:ms', fEnd - fStart));
end;

constructor TDelta.Create;
begin
  vSettings := TSettings.Create;
  fNewDataSet := TMemDataset.Create(nil);
  fOldDataSet := TMemDataset.Create(nil);
end;

destructor TDelta.Destroy;
begin
  vSettings.Free;
  fNewDataSet.Free;
  fOldDataSet.Free;
  inherited Destroy;
end;

procedure TDelta.Execute(ANewFile, AOldFile, ADeltaFile: string; showDebugOption: integer);
begin
  fNewFile := ANewFile;
  fOldFile := AOldFile;
  fDeltaFile := ADeltaFile;
  fstart := Now;
  if FileExists(DeltaFile) then
    DeleteFile(DeltaFile);

  fDebugOption := showDebugOption;
  //startCompare(showDebugOption);
  CreateDataSet;
  if not fNewDataSet.Active then
    fNewDataSet.Open;
  if not fOldDataSet.Active then
    fOldDataSet.Open;
  AddToDataSet(NewFile, True);
  AddToDataSet(OldFile, False);
  WriteFileDetails;
  //fNewDataSet.SaveToFile('F:\WORK\Tech Solutions\claimload_synch\load_delta\test.csv');
  compareDelta;
  //compareStart;
end;

{ TSettings }

constructor TSettings.Create;
var
  afile: string;
begin
  afile := ExtractFilePath(ParamStr(0)) + 'load_delta.conf';
  fini := TInifile.Create(afile);
end;

destructor TSettings.Destroy;
begin
  fini.Free;
  inherited Destroy;
end;

function TSettings.sectionValue(ASection, AIndent: string): string;
begin
  if not FileExists(fini.FileName) then
  begin
    WriteLn('Configuration file does not exists');
    Result := '';
    Exit;
  end;
  Result := fini.ReadString(ASection, AIndent, '');
end;

end.
